// Copyright (c) 2010, Google Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// linux_dumper.cc: Implement google_breakpad::LinuxDumper.
// See linux_dumper.h for details.

// This code deals with the mechanics of getting information about a crashed
// process. Since this code may run in a compromised address space, the same
// rules apply as detailed at the top of minidump_writer.h: no libc calls and
// use the alternative allocator.

#include "client/linux/minidump_writer/linux_dumper.h"

#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <stddef.h>
#include <string.h>
#include <sys/types.h>

#include "client/linux/minidump_writer/line_reader.h"
#include "common/linux/file_id.h"
#include "common/linux/linux_libc_support.h"
#include "common/linux/memory_mapped_file.h"
#include "common/linux/safe_readlink.h"
#include "third_party/lss/linux_syscall_support.h"

static const char kMappedFileUnsafePrefix[] = "/dev/";
static const char kDeletedSuffix[] = " (deleted)";

inline static bool IsMappedFileOpenUnsafe(
    const google_breakpad::MappingInfo& mapping) {
  // It is unsafe to attempt to open a mapped file that lives under /dev,
  // because the semantics of the open may be driver-specific so we'd risk
  // hanging the crash dumper. And a file in /dev/ almost certainly has no
  // ELF file identifier anyways.
  return my_strncmp(mapping.name,
                    kMappedFileUnsafePrefix,
                    sizeof(kMappedFileUnsafePrefix) - 1) == 0;
}

namespace google_breakpad {

// All interesting auvx entry types are below AT_SYSINFO_EHDR
#define AT_MAX AT_SYSINFO_EHDR

LinuxDumper::LinuxDumper(pid_t pid)
    : pid_(pid),
      crash_address_(0),
      crash_signal_(0),
      crash_thread_(pid),
      threads_(&allocator_, 8),
      mappings_(&allocator_),
      auxv_(&allocator_, AT_MAX + 1),
      executable_device_(0),
      executable_inode_(0) {
  // The passed-in size to the constructor (above) is only a hint.
  // Must call .resize() to do actual initialization of the elements.
  auxv_.resize(AT_MAX + 1);
}

LinuxDumper::~LinuxDumper() {
}

bool LinuxDumper::Init() {
  return ReadAuxv() && EnumerateThreads() && EnumerateMappings();
}

bool
LinuxDumper::ElfFileIdentifierForMapping(const MappingInfo& mapping,
                                         uint8_t identifier[sizeof(MDGUID)]) {
  my_memset(identifier, 0, sizeof(MDGUID));
  if (IsMappedFileOpenUnsafe(mapping))
    return false;

  // Special-case linux-gate because it's not a real file.
  if (mapping.start_addr == auxv_[AT_SYSINFO_EHDR]) {
    MemoryRange linux_gate;
    if (pid_ == sys_getpid()) {
      linux_gate.Set(reinterpret_cast<void*>(mapping.start_addr), mapping.size);
    } else {
      linux_gate.Set(allocator_.Alloc(mapping.size), mapping.size);
      CopyFromProcess(const_cast<uint8_t*>(linux_gate.data()), pid_,
                      reinterpret_cast<const void*>(mapping.start_addr),
                      mapping.size);
    }
    return FileID::ElfFileIdentifierFromMappedFile(linux_gate, identifier);
  }

  MemoryMappedFile mapped_file;
  if (mapping.deleted) {
    char filename[NAME_MAX];
    if (!executable_device_ && !executable_inode_) {
      if (!BuildProcPath(filename, pid_, "exe"))
        return false;
      struct kernel_stat stat;
      if (sys_stat(filename, &stat) != 0)
        return false;
      executable_device_ = stat.st_dev;
      executable_inode_ = stat.st_ino;
    }
    if (mapping.device == executable_device_ &&
        mapping.inode == executable_inode_) {
    }
    mapped_file.Map(filename);
  } else {
    mapped_file.Map(mapping.name);
  }

  if (!mapped_file.data())  // Should probably check if size >= ElfW(Ehdr)?
    return false;

  return
      FileID::ElfFileIdentifierFromMappedFile(mapped_file.content(), identifier);
}

bool LinuxDumper::ReadAuxv() {
  char auxv_path[NAME_MAX];
  if (!BuildProcPath(auxv_path, pid_, "auxv")) {
    return false;
  }

  int fd = sys_open(auxv_path, O_RDONLY, 0);
  if (fd < 0) {
    return false;
  }

  elf_aux_entry one_aux_entry;
  bool res = false;
  while (sys_read(fd,
                  &one_aux_entry,
                  sizeof(elf_aux_entry)) == sizeof(elf_aux_entry) &&
         one_aux_entry.a_type != AT_NULL) {
    if (one_aux_entry.a_type <= AT_MAX) {
      auxv_[one_aux_entry.a_type] = one_aux_entry.a_un.a_val;
      res = true;
    }
  }
  sys_close(fd);
  return res;
}

bool LinuxDumper::EnumerateMappings() {
  static const size_t kDeletedSuffixLen = sizeof(kDeletedSuffix) - 1;

  char maps_path[NAME_MAX];
  if (!BuildProcPath(maps_path, pid_, "maps"))
    return false;

  // Although the initial executable is usually the first mapping, it's not
  // guaranteed (see http://crosbug.com/25355); therefore, try to use the
  // actual entry point to find the mapping.
  const void* entry_point_loc = reinterpret_cast<void *>(auxv_[AT_ENTRY]);

  const int fd = sys_open(maps_path, O_RDONLY, 0);
  if (fd < 0)
    return false;
  LineReader* const line_reader = new(allocator_) LineReader(fd);

  const char* line;
  unsigned line_len;
  while (line_reader->GetNextLine(&line, &line_len)) {
    uintptr_t start_addr, end_addr, offset;
    uintptr_t dev_major, dev_minor, inode;

    const char* i = my_read_hex_ptr(&start_addr, line);
    if (*i != '-')
      goto popline;
    i = my_read_hex_ptr(&end_addr, i + 1);
    if (*i != ' ')
      goto popline;
    i = my_read_hex_ptr(&offset, i + 6 /* skip ' rwxp ' */);
    if (*i != ' ')
      goto popline;
    i = my_read_hex_ptr(&dev_major, i + 1);
    if (*i != ':')
      goto popline;
    i = my_read_hex_ptr(&dev_minor, i + 1);
    if (*i != ' ')
      goto popline;
    i = my_read_decimal_ptr(&inode, i + 1);
    if (*i != ' ')
      goto popline;

    dev_t device;
    device = makedev(dev_major, dev_minor);
    const char* name;
    // Skip whitespace.
    for (name = i; *name == ' '; name++);
    if (!*name)
      name = NULL;
    // Merge adjacent mappings with the same device/inode pair into one module,
    // assuming they're a single library mapped by the dynamic linker
    if (name && !mappings_.empty()) {
      MappingInfo* module = mappings_.back();
      if ((start_addr == module->start_addr + module->size) &&
          (device == module->device) && (inode == module->inode)) {
        module->size = end_addr - module->start_addr;
        line_reader->PopLine(line_len);
        continue;
      }
    }
    MappingInfo* module;
    module = new(allocator_) MappingInfo;
    my_memset(module, 0, sizeof(MappingInfo));
    module->start_addr = start_addr;
    module->size = end_addr - start_addr;
    module->offset = offset;
    module->device = device;
    module->inode = inode;
    if (name != NULL) {
      unsigned l = my_strlen(name);
      // Check for ' (deleted)' in |name|.
      // |name| has to be at least as long as "/x (deleted)".
      if ((l >= kDeletedSuffixLen + 2) &&
          (my_strncmp(name + l - kDeletedSuffixLen, kDeletedSuffix,
                      kDeletedSuffixLen) == 0)) {
        // Check to see if someone actually named a file 'foo (deleted)'.
        struct kernel_stat stat;
        if (sys_stat(name, &stat) != 0 ||
            stat.st_ino != inode || stat.st_dev != device) {
          module->deleted = true;
          l -= kDeletedSuffixLen;
        }
      }

      if (l < sizeof(module->name))
        my_memcpy(module->name, name, l);
    }
    // If this is the entry-point mapping, and it's not already the
    // first one, then we need to make it be first.  This is because
    // the minidump format assumes the first module is the one that
    // corresponds to the main executable (as codified in
    // processor/minidump.cc:MinidumpModuleList::GetMainModule()).
    if (entry_point_loc &&
        (entry_point_loc >=
            reinterpret_cast<void*>(module->start_addr)) &&
        (entry_point_loc <
            reinterpret_cast<void*>(module->start_addr+module->size)) &&
        !mappings_.empty()) {
      // push the module onto the front of the list.
      mappings_.resize(mappings_.size() + 1);
      for (size_t idx = mappings_.size() - 1; idx > 0; idx--)
        mappings_[idx] = mappings_[idx - 1];
      mappings_[0] = module;
    } else {
      mappings_.push_back(module);
    }
popline:
    line_reader->PopLine(line_len);
  }

  sys_close(fd);

  return !mappings_.empty();
}

// Get information about the stack, given the stack pointer. We don't try to
// walk the stack since we might not have all the information needed to do
// unwind. So we just grab, up to, 32k of stack.
bool LinuxDumper::GetStackInfo(const void** stack, size_t* stack_len,
                               uintptr_t int_stack_pointer) {
  // Move the stack pointer to the bottom of the page that it's in.
  const uintptr_t page_size = getpagesize();

  uint8_t* const stack_pointer =
      reinterpret_cast<uint8_t*>(int_stack_pointer & ~(page_size - 1));

  // The number of bytes of stack which we try to capture.
  static const ptrdiff_t kStackToCapture = 32 * 1024;

  const MappingInfo* mapping = FindMapping(stack_pointer);
  if (!mapping)
    return false;
  const ptrdiff_t offset = stack_pointer -
      reinterpret_cast<uint8_t*>(mapping->start_addr);
  const ptrdiff_t distance_to_end =
      static_cast<ptrdiff_t>(mapping->size) - offset;
  *stack_len = distance_to_end > kStackToCapture ?
      kStackToCapture : distance_to_end;
  *stack = stack_pointer;
  return true;
}

// Find the mapping which the given memory address falls in.
const MappingInfo* LinuxDumper::FindMapping(const void* address) const {
  const uintptr_t addr = (uintptr_t) address;

  for (size_t i = 0; i < mappings_.size(); ++i) {
    const uintptr_t start = static_cast<uintptr_t>(mappings_[i]->start_addr);
    if (addr >= start && addr - start < mappings_[i]->size)
      return mappings_[i];
  }

  return NULL;
}

}  // namespace google_breakpad
